/*
 * A flexpage abstraction.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/dataspace.h>

#include <algorithm>

#include "flexpage.h"



/* Reset the flexpage using 'offset', being the file offset. */

void Flexpage::reset(offset_t offset)
{
    _counter = 0;
    _flags = 0;

    /* By definition (see "Flexible-Sized Page Objects - Object-Orientation
       in Operation Systems"), flexpages are aligned to multiples of their
       size.

       The size of the flexpage depends on the amount of space around the
       accessed page. It cannot exceed the size of the memory region. */

    size = max_multiple(region->start, region->end, PAGE_SIZE);

    /* The base address of the flexpage is computed from the region start
       and flexpage size. It will be no lower than the region start.

       Sent flexpages may use higher bases due to receive window constraints,
       these being communicated by the "hot spot". */

    base_addr = round(region->start, size);

    /* Get the file offset for the base of the flexpage. This will be a
       multiple of the flexpage size for alignment purposes. */

    base_offset = trunc(offset, size);

    /* The page being accessed is relative to the base.
       (This is transient information recording the initialising access
       details.) */

    page_offset = trunc(offset - base_offset, PAGE_SIZE);
    page_addr = base_addr + page_offset;
}

/* Decrement the usage counter, returning whether the flexpage is now no longer
   used. */

bool Flexpage::decrement()
{
    if (_counter)
    {
        _counter--;
        return _counter == 0;
    }
    else
        return 0;
}

/* Increment the usage counter. */

void Flexpage::increment()
{
    _counter++;
}

/* Invalidate the flexpage, meaning that it should not now be in use. */

void Flexpage::invalidate()
{
    _counter = 0;
}

/* Return whether the flexpage is in use and is therefore valid. */

bool Flexpage::valid()
{
    return _counter != 0;
}

/* Return whether the flexpage supports the given file 'position'. */

bool Flexpage::supports_position(offset_t position)
{
    return (base_offset <= position) && (position < (base_offset + size));
}

/* Upgrade the flags involved with this flexpage. This is used to track the
   maximal flags employed by the different pagers, with the result being used in
   unmap operations. */

void Flexpage::upgrade(map_flags_t flags)
{
    if (flags && (flags != _flags))
        _flags |= flags;
}

/* Return whether the flexpage has been modified due to write access having been
   granted for any user of the page. */

bool Flexpage::modified()
{
    return _flags & L4RE_DS_F_W;
}

/* Return a "send" flexpage for an access to 'offset' by positioning it relative
   to 'hot_spot' for the receive flexpage window. */

SendFlexpage Flexpage::to_send(offset_t offset, offset_t hot_spot,
                               map_flags_t flags, offset_t max_offset)
{
    /* The dataspace offset of the flexpage base is a multiple of the flexpage
       size. */

    offset_t receive_base_offset = trunc(offset, size);

    /* The offset of the accessed page within the flexpage is constrained by the
       current flexpage size. */

    offset_t page_offset = trunc(offset - receive_base_offset, PAGE_SIZE);

    /* The receive flexpage offset (hot spot) must be constrained to the
       flexpage, both the size and the start. */

    offset_t receive_size;

    if (max_offset)
    {
        receive_size = trunc_multiple(max_offset - receive_base_offset, PAGE_SIZE);
        receive_size = std::min(size, receive_size);
    }
    else
        receive_size = size;

    if (!receive_size)
        return SendFlexpage(base_addr, page_order(0), flags);

    offset_t receive_page_offset = hot_spot % receive_size;

    while ((receive_size > PAGE_SIZE) && (receive_page_offset != page_offset))
    {
        receive_size /= 2;
        receive_page_offset = hot_spot % receive_size;
    }

    /* The flexpage base address is adjusted using the difference in page
       offsets. Where the receive flexpage offset is constained further, the
       base address will be raised to become closer to the accessed page. */

    offset_t adjustment = page_offset - receive_page_offset;

    return SendFlexpage(base_addr + adjustment, page_order(receive_size), flags);
}

/* Return a representation of the flexpage for unmapping. */

SendFlexpage Flexpage::to_unmap()
{
    return SendFlexpage(base_addr, page_order(size), _flags);
}

// vim: tabstop=4 expandtab shiftwidth=4
